#include "globals.h"

/*  default bitmaps  */
#include "bitmaps/block_special.xbm"
#include "bitmaps/blank.xbm"
#include "bitmaps/char_special.xbm"

#ifdef HAVE_XPMLIB
#include <X11/xpm.h>
#include "bitmaps/exec.xpm"
#include "bitmaps/house.xpm"
#include "bitmaps/multiple.xpm"
#include "bitmaps/network.xpm"
#include "bitmaps/plain_doc.xpm"
#include "bitmaps/plain_folder.xpm"
#include "bitmaps/link_dir.xpm"
#include "bitmaps/duplicate.xpm"
#include "bitmaps/delete.xpm"
#include "bitmaps/attributes.xpm"
#include "bitmaps/createDir.xpm"
#include "bitmaps/ftp.xpm"
#include "bitmaps/filing.xpm"
#include "bitmaps/question.xpm"
#include "bitmaps/floppy.xpm"
#include "bitmaps/printer.xpm"
#include "bitmaps/tapes.xpm"
#else
#include "bitmaps/exec.xbm"
#include "bitmaps/house.xbm"
#include "bitmaps/multiple.xbm"
#include "bitmaps/network.xbm"
#include "bitmaps/plain_doc.xbm"
#include "bitmaps/plain_folder.xbm"
#include "bitmaps/link_dir.xbm"
#include "bitmaps/duplicate.xbm"
#include "bitmaps/delete.xbm"
#include "bitmaps/attributes.xbm"
#include "bitmaps/createDir.xbm"
#include "bitmaps/ftp.xbm"
#include "bitmaps/filing32.xbm"
#include "bitmaps/question.xbm"
#include "bitmaps/floppy.xbm"
#include "bitmaps/printer.xbm"
#include "bitmaps/tapes.xbm"
#endif

#define XtNdefaultLaunchCommand         "defaultLaunchCommand"
#define XtCDefaultLaunchCommand         "DefaultLaunchCommand"
#define XtNrules			"rules"
#define XtCRules			"Rules"
#define XtNfilters			"filters"
#define XtCFilters			"Filters"

#ifndef XBMPATH
#define XBMPATH  "/usr/include/X11/bitmaps/"
#endif

#define APP_RES_FILENAME		".mfmrc"

/*
    order:  
	1. is multiple files?
	2. by filename only (ex. core, Makefile)
	3. by user-specified extension (ex. .xbm, .ps)
	4. by 'file' command
	5. by IS_... including IS_EXECUTABLE
*/


typedef struct
{
  String defaultLaunchCommand, 
	 rules,
	 filters;
} AppResData;

typedef struct SpecialTypes
{
  Pixmap DIRECTORY, SOCKET, FIFO, CHAR_SPECIAL, HOME,
         BLK_SPECIAL, REG_DEFAULT, EXECUTABLE, MULTIPLE,
	 LINK, BLANK, PRINTER, FLOPPY, TAPES;
} SpecialTypes;

static XtResource resources[] = {
  {
        XtNdefaultLaunchCommand,
        XtCDefaultLaunchCommand,
        XtRString,
        sizeof( String ),
        XtOffsetOf( AppResData, defaultLaunchCommand ),
        XtRString,
        (XtPointer)""
  },
  {
        XtNrules,
        XtCRules,
        XtRString,
        sizeof( String ),
        XtOffsetOf( AppResData, rules ),
        XtRString,
        (XtPointer)""
  },
  {
	XtNfilters,
	XtCFilters,
	XtRString,
        sizeof( String ),
        XtOffsetOf( AppResData, filters ),
        XtRString,
        (XtPointer)""
  }
};

static int numRuleTypes = 6;
static char *ruleTypes[] = { 
		"FILENAME", 
		"FILE+COMM",
		"EXTENSION",
		"EXT+COMMAND", 
		"FILESTRING", 
		"FILESTR+COMM",
		};

#define _FILENAME 		0
#define _FILE_COMM		1
#define _EXTENSION		2
#define _EXT_COMMAND		3
#define _FILESTRING		4
#define _FILESTR_COMM		5

#define _NUM_RULE_TYPES		6

static AppResData appResData;
static char defaultCommand[256], editorCommand[256];
static SpecialTypes specialTypes;
static struct Rule *rules = (struct Rule *)NULL;

static Boolean isPoppedUp = False;
static Widget appResShell = NULL, rulesPB[_NUM_RULE_TYPES], defaultCmdTF, devSL,
	      devLabelTF, devPixmapTF, devCmdTF, rulePixmapTF, ruleSpecTF, 
	      ruleCmdTF, ruleSL, currentRuleType, ruleLabel1, ruleLabel2, 
	      bpChooser = (Widget)NULL;
static int currentRuleTypeIndex = 0;


void bpChooserOKCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void bpChooserApplyCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void bpChooserCancelCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void popupBPChooser(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void loadRules(
#if NeedFunctionPrototypes
	Widget, int
#endif
);
void insertDevCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void deleteDevCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void insertRuleCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void deleteRuleCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void ruleTypeCB(
#if NeedFunctionPrototypes
	Widget, XtPointer, XtPointer
#endif
);
void loadIncludedPixmap(
#if NeedFunctionPrototypes
  Display*, Window, char**, Pixmap*, Pixel, char*
#endif
);
Pixmap loadPixmap( 
#if NeedFunctionPrototypes
	Widget, char*, Pixel, Pixel 
#endif
);
Boolean extensionMatches( 
#if NeedFunctionPrototypes
	char*, char* 
#endif
);
void loadSpecialPixmaps( 
#if NeedFunctionPrototypes
	Widget, Pixel, Pixel 
#endif
);
int fileCmd( 
#if NeedFunctionPrototypes
	char*, char* 
#endif
);
void parseRules( 
#if NeedFunctionPrototypes
  char*, Widget, Pixel, Pixel
#endif
);
void appResApplyCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void appResSaveCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void appResDoneCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);


void configCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  if( !appResShell )
  {
    XFontStruct *font;
    XmFontList  fontList;
    int i, n_args;
    Arg args[15];
    XmString xms;
    Atom wmDel;
    Widget appResTopForm, separator, applyButton, saveButton, doneButton,
	   appResForm, defaultCmdFrame, defaultCmdFrameLabel, defaultCmdForm,
	   rulesFrame, rulesForm, rulesPM, rulesOM, devFrame, devFrameLabel,
	   devForm, devLabel, devPB, rulePB, ruleLabel, msgForm, msgLabel1,
	   msgLabel2, msgLabel3, msgLabel4;

    appResShell = XtVaCreatePopupShell( "Configuration",
	transientShellWidgetClass, w,
	XmNallowShellResize, False,
	NULL );

    appResTopForm = XtVaCreateManagedWidget( "appResTopForm",
	xmFormWidgetClass, appResShell, NULL );

    XtVaSetValues( appResShell, XmNdeleteResponse, XmDO_NOTHING, NULL );
    wmDel = XmInternAtom( XtDisplay(appResTopForm), "WM_DELETE_WINDOW", False );
    XmAddWMProtocolCallback( appResShell, wmDel, appResDoneCB, appResTopForm );

    separator = XtVaCreateManagedWidget( "separator",
                xmSeparatorWidgetClass, appResTopForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 60,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                NULL );

    applyButton = XtVaCreateManagedWidget( "Apply",
                xmPushButtonWidgetClass, appResTopForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 10,
                XmNleftAttachment, XmATTACH_POSITION,
                XmNleftPosition, 7,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 31,
                XmNheight, 30,
                XmNshowAsDefault, True,
                NULL );
    XtAddCallback( applyButton, XmNactivateCallback, appResApplyCB, NULL );

    saveButton = XtVaCreateManagedWidget( "Save",
                xmPushButtonWidgetClass, appResTopForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 15,
                XmNleftAttachment, XmATTACH_POSITION,
                XmNleftPosition, 38,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 62,
                XmNheight, 30,
                NULL );
    XtAddCallback( saveButton, XmNactivateCallback, appResSaveCB, NULL );

    doneButton = XtVaCreateManagedWidget( "Done",
		xmPushButtonWidgetClass, appResTopForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 15,
                XmNleftAttachment, XmATTACH_POSITION,
                XmNleftPosition, 69,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 93,
                XmNheight, 30,
                NULL );
    XtAddCallback( doneButton, XmNactivateCallback, appResDoneCB, NULL );

    appResForm = XtVaCreateManagedWidget( "propertiesForm",
                xmFormWidgetClass, appResTopForm,
                XmNtopAttachment, XmATTACH_FORM,
                XmNtopOffset, 10,
                XmNleftAttachment, XmATTACH_FORM,
                XmNleftOffset, 10,
                XmNrightAttachment, XmATTACH_FORM,
                XmNrightOffset, 10,
                XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, separator,
                XmNbottomOffset, 10,
                NULL );

    msgForm = XtVaCreateManagedWidget( "msgForm",
		xmFormWidgetClass, appResForm,
		XmNtopAttachment, XmATTACH_FORM,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNbottomAttachment, XmATTACH_OPPOSITE_FORM,
                XmNbottomOffset, -225,
                NULL );
    xms = XmStringCreateLtoR( "When creating commands, keep in mind that the following macros are predefined:",
	XmSTRING_DEFAULT_CHARSET );
    msgLabel1 = XtVaCreateManagedWidget( "msgLabel1",
		xmLabelWidgetClass, msgForm,
		XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_FORM,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNbottomAttachment, XmATTACH_OPPOSITE_FORM,
                XmNbottomOffset, -40,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateSimple( "$PATH+FILENAME" );
    msgLabel2 = XtVaCreateManagedWidget( "msgLabel2",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, msgLabel1,
    		XmNtopOffset, 5,
                XmNrightAttachment, XmATTACH_POSITION,
		XmNrightPosition, 30,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateLtoR( "The complete filename including path, i.e., /usr/renalds/.cshrc", 
	XmSTRING_DEFAULT_CHARSET );
    msgLabel3 = XtVaCreateManagedWidget( "msgLabel3",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
		XmNalignment, XmALIGNMENT_BEGINNING,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel1,
                XmNtopOffset, 5,
                XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, msgLabel2,
		XmNleftOffset, 35,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateSimple( "$FILENAME" );
    msgLabel1 = XtVaCreateManagedWidget( "msgLabel1",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel2,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 30,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateLtoR( "The filename excluding path, i.e., .cshrc",
        XmSTRING_DEFAULT_CHARSET );
    msgLabel4 = XtVaCreateManagedWidget( "msgLabel4",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
		XmNalignment, XmALIGNMENT_BEGINNING,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel3,
                XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, msgLabel1,
		XmNleftOffset, 35,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateSimple( "$PATH" );
    msgLabel2 = XtVaCreateManagedWidget( "msgLabel2",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel1,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 30,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateLtoR( "The path of the file only, i.e., /usr/renalds",
        XmSTRING_DEFAULT_CHARSET );
    msgLabel3 = XtVaCreateManagedWidget( "msgLabel3",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
		XmNalignment, XmALIGNMENT_BEGINNING,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel4,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, msgLabel2,
		XmNleftOffset, 35,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateSimple( "$EDITOR" );
    msgLabel1 = XtVaCreateManagedWidget( "msgLabel1",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel2,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 30,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateLtoR( "The environment variable EDITOR (if set to vi, mfm will execute it in an xterm)",
        XmSTRING_DEFAULT_CHARSET );
    msgLabel4 = XtVaCreateManagedWidget( "msgLabel4",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
		XmNalignment, XmALIGNMENT_BEGINNING,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel3,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, msgLabel1,
		XmNleftOffset, 35,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateSimple( "$ALL" );
    msgLabel2 = XtVaCreateManagedWidget( "msgLabel2",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel1,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 30,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateLtoR( "When multiple files are dropped onto a device drop site, the command can be executed\nonce for each file name by writing a command such as \"lp $PATH+FILENAME\" or the command\ncan be executed once including all of the file names by writing the same command as \"lp $ALL\"", XmSTRING_DEFAULT_CHARSET );
    msgLabel3 = XtVaCreateManagedWidget( "msgLabel3",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
		XmNalignment, XmALIGNMENT_BEGINNING,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, msgLabel4,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, msgLabel2,
                XmNleftOffset, 35,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateLtoR( "The following device pixmaps are also predefined for your convenience:  \"PRINTER\", \"FLOPPY\", \"TAPES\"", XmSTRING_DEFAULT_CHARSET );
    msgLabel1 = XtVaCreateManagedWidget( "msgLabel1",
                xmLabelWidgetClass, msgForm,
                XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, msgLabel3,
		XmNtopOffset, 10,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
		XmNheight, 30,
                NULL );
    XmStringFree( xms );
    xms = XmStringCreateLtoR( "", XmSTRING_DEFAULT_CHARSET );
    msgLabel2 = XtVaCreateManagedWidget( "msgLabel2",
		xmLabelWidgetClass, msgForm,
		NULL );
    XmStringFree( xms );

    defaultCmdFrame = XtVaCreateManagedWidget( "defaultCmdFrame",
                xmFrameWidgetClass, appResForm,
                XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, msgForm,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET,
		XmNbottomWidget, msgForm,
                XmNbottomOffset, -65,
                XmNshadowType, XmSHADOW_OUT,
                XmNshadowThickness, 3,
                NULL );

    defaultCmdFrameLabel = XtVaCreateManagedWidget( "Default Command",
                xmLabelWidgetClass, defaultCmdFrame,
                XmNchildHorizontalAlignment, XmALIGNMENT_CENTER,
                XmNchildType, XmFRAME_TITLE_CHILD,
                NULL );

    defaultCmdForm = XtVaCreateManagedWidget( "defaultCmdForm",
                xmFormWidgetClass, defaultCmdFrame,
                XmNchildType, XmFRAME_WORKAREA_CHILD,
                NULL );

    defaultCmdTF = XtVaCreateManagedWidget( "defaultCmdTF",
		xmTextFieldWidgetClass, defaultCmdForm,
		XmNbottomAttachment, XmATTACH_FORM,
		XmNbottomOffset, 5,
		XmNleftAttachment, XmATTACH_FORM,
		XmNleftOffset, 10,
		XmNrightAttachment, XmATTACH_FORM,
		XmNrightOffset, 10,
		XmNheight, 30,
		XmNmarginHeight, 0,
		XmNeditable, True,
		XmNwidth, 1000,
		NULL );

    devFrame = XtVaCreateManagedWidget( "devFrame",
                xmFrameWidgetClass, appResForm,
                XmNtopAttachment, XmATTACH_OPPOSITE_FORM,
                XmNtopOffset, -200,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNshadowType, XmSHADOW_OUT,
                XmNshadowThickness, 3,
                NULL );

    devFrameLabel = XtVaCreateManagedWidget( "Devices",
                xmLabelWidgetClass, devFrame,
                XmNchildHorizontalAlignment, XmALIGNMENT_CENTER,
                XmNchildType, XmFRAME_TITLE_CHILD,
                NULL );

    devForm = XtVaCreateManagedWidget( "devForm",
                xmFormWidgetClass, devFrame,
                XmNchildType, XmFRAME_WORKAREA_CHILD,
                NULL );

    font = XLoadQueryFont( XtDisplay(w), "*misc*fixed*bold*14*" );
    fontList = XmFontListCreate( font, "set1" );

    n_args = 0;
    XtSetArg( args[n_args], XmNtopAttachment, XmATTACH_FORM );  n_args++;
    XtSetArg( args[n_args], XmNtopOffset, 35 );  n_args++;
    XtSetArg( args[n_args], XmNleftAttachment, XmATTACH_FORM ); n_args++;
    XtSetArg( args[n_args], XmNleftOffset, 10 );  n_args++;
    XtSetArg( args[n_args], XmNrightAttachment, XmATTACH_FORM ); n_args++;
    XtSetArg( args[n_args], XmNrightOffset, 10 );  n_args++;
    XtSetArg( args[n_args], XmNbottomAttachment, XmATTACH_FORM ); n_args++;
    XtSetArg( args[n_args], XmNbottomOffset, 25 );  n_args++;
    XtSetArg( args[n_args], XmNshadowThickness, 3 );  n_args++;
    XtSetArg( args[n_args], XmNscrollBarDisplayPolicy, XmSTATIC );  n_args++;
    XtSetArg( args[n_args], XmNselectionPolicy, XmSINGLE_SELECT );  n_args++;
    XtSetArg( args[n_args], XmNfontList, fontList );  n_args++;
    devSL = XmCreateScrolledList( devForm, "devSL", args, n_args );
    XtManageChild( devSL );

    devPB = XtVaCreateManagedWidget( "Insert",
		xmPushButtonWidgetClass, devForm,
		XmNbottomAttachment, XmATTACH_FORM,
		XmNleftAttachment, XmATTACH_FORM,
		XmNleftOffset, 10,
		XmNrightAttachment, XmATTACH_POSITION,
		XmNrightPosition, 50,
		XmNheight, 25,
		NULL );
    XtAddCallback( devPB, XmNactivateCallback, insertDevCB, NULL );
    devPB = XtVaCreateManagedWidget( "Delete",
                xmPushButtonWidgetClass, devForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNrightOffset, 10,
                XmNleftAttachment, XmATTACH_POSITION,
                XmNleftPosition, 50,
		XmNheight, 25,
                NULL );
    XtAddCallback( devPB, XmNactivateCallback, deleteDevCB, NULL );
    devLabel = XtVaCreateManagedWidget( "Label String: ",
		xmLabelWidgetClass, devForm,
		XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, devSL,
		XmNheight, 30,
		XmNleftAttachment, XmATTACH_FORM,
		XmNleftOffset, 10,
		NULL );
    devLabelTF = XtVaCreateManagedWidget( "devLabelTF",
		xmTextFieldWidgetClass, devForm,
		XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, devSL,
                XmNheight, 30,
		XmNwidth, 125,
                XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, devLabel,
		XmNmarginHeight, 0,
		NULL );
    devLabel = XtVaCreateManagedWidget( "Pixmap: ",
                xmPushButtonWidgetClass, devForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, devSL,
                XmNheight, 30,
		XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, devLabelTF,
                NULL );
    devPixmapTF = XtVaCreateManagedWidget( "devPixmapTF",
                xmTextFieldWidgetClass, devForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, devSL,
                XmNheight, 30,
		XmNwidth, 250,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, devLabel,
		XmNmarginHeight, 0,
                NULL );
    XtAddCallback( devLabel, XmNactivateCallback, popupBPChooser, 
        (XtPointer)devPixmapTF );
    devLabel = XtVaCreateManagedWidget( "Command: ",
                xmLabelWidgetClass, devForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, devSL,
                XmNheight, 30,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, devPixmapTF,
                NULL );
    devCmdTF = XtVaCreateManagedWidget( "devPixmapTF",
                xmTextFieldWidgetClass, devForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, devSL,
                XmNheight, 30,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, devLabel,
		XmNrightAttachment, XmATTACH_FORM,
		XmNrightOffset, 10,
		XmNmarginHeight, 0,
                NULL );

    rulesFrame = XtVaCreateManagedWidget( "defaultCmdFrame",
                xmFrameWidgetClass, appResForm,
                XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, defaultCmdFrame,
		XmNtopOffset, 10,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNbottomAttachment, XmATTACH_WIDGET,
		XmNbottomWidget, devFrame,
		XmNbottomOffset, 10,
                XmNshadowType, XmSHADOW_OUT,
                XmNshadowThickness, 3,
                NULL );

    rulesPM = XmCreatePulldownMenu( rulesFrame, "rulesPM", NULL, 0 );
    for( i = 0; i < numRuleTypes; i++ )
    {
      rulesPB[i] = XtVaCreateManagedWidget( ruleTypes[i],
                xmPushButtonWidgetClass, rulesPM, NULL, 0 );
      XtAddCallback( rulesPB[i], XmNactivateCallback, ruleTypeCB, NULL );
    }
    currentRuleTypeIndex = 0;

    xms = XmStringCreateSimple( "Rule Types: " );
    n_args = 0;
    XtSetArg( args[n_args], XmNsubMenuId, rulesPM );  n_args++;
    XtSetArg( args[n_args], XmNchildType, XmFRAME_TITLE_CHILD );  n_args++;
    XtSetArg( args[n_args], XmNchildHorizontalAlignment, XmALIGNMENT_CENTER );  n_args++;
    XtSetArg( args[n_args], XmNlabelString, xms );  n_args++;
    rulesOM = XmCreateOptionMenu( rulesFrame, "Rule Types: ", args, n_args );
    XtManageChild( rulesOM );
    XmStringFree( xms );

    rulesForm = XtVaCreateManagedWidget( "srvForm",
                xmFormWidgetClass, rulesFrame,
                XmNchildType, XmFRAME_WORKAREA_CHILD,
                NULL );

    n_args = 0;
    XtSetArg( args[n_args], XmNtopAttachment, XmATTACH_FORM );  n_args++;
    XtSetArg( args[n_args], XmNtopOffset, 35 );  n_args++;
    XtSetArg( args[n_args], XmNleftAttachment, XmATTACH_FORM ); n_args++;
    XtSetArg( args[n_args], XmNleftOffset, 10 );  n_args++;
    XtSetArg( args[n_args], XmNrightAttachment, XmATTACH_FORM ); n_args++;
    XtSetArg( args[n_args], XmNrightOffset, 10 );  n_args++;
    XtSetArg( args[n_args], XmNbottomAttachment, XmATTACH_FORM ); n_args++;
    XtSetArg( args[n_args], XmNbottomOffset, 25 );  n_args++;
    XtSetArg( args[n_args], XmNshadowThickness, 3 );  n_args++;
    XtSetArg( args[n_args], XmNscrollBarDisplayPolicy, XmSTATIC );  n_args++;
    XtSetArg( args[n_args], XmNselectionPolicy, XmSINGLE_SELECT );  n_args++;
    XtSetArg( args[n_args], XmNvisibleItemCount, 10 );  n_args++;
    XtSetArg( args[n_args], XmNfontList, fontList );  n_args++;
    ruleSL = XmCreateScrolledList( rulesForm, "ruleSL", args, n_args );
    XtManageChild( ruleSL );

    rulePB = XtVaCreateManagedWidget( "Insert",
                xmPushButtonWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNleftAttachment, XmATTACH_FORM,
                XmNleftOffset, 10,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 50,
                XmNheight, 25,
                NULL );
    XtAddCallback( rulePB, XmNactivateCallback, insertRuleCB, NULL );
    rulePB = XtVaCreateManagedWidget( "Delete",
                xmPushButtonWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNrightOffset, 10,
                XmNleftAttachment, XmATTACH_POSITION,
                XmNleftPosition, 50,
                XmNheight, 25,
                NULL );
    XtAddCallback( rulePB, XmNactivateCallback, deleteRuleCB, NULL );
    ruleLabel = XtVaCreateManagedWidget( "Pixmap: ",
                xmPushButtonWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, ruleSL,
                XmNheight, 30,
		XmNleftAttachment, XmATTACH_FORM,
                XmNleftOffset, 10,
                NULL );
    rulePixmapTF = XtVaCreateManagedWidget( "rulePixmapTF",
                xmTextFieldWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, ruleSL,
                XmNheight, 30,
                XmNwidth, 250,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, ruleLabel,
                XmNmarginHeight, 0,
                NULL );
    XtAddCallback( ruleLabel, XmNactivateCallback, popupBPChooser,
        (XtPointer)rulePixmapTF );
    ruleLabel1 = XtVaCreateManagedWidget( "Filename: ",
		xmLabelWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, ruleSL,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, rulePixmapTF,
		XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
		XmNtopWidget, rulePixmapTF,
                NULL );
    ruleSpecTF = XtVaCreateManagedWidget( "ruleSpecTF",
		xmTextFieldWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, ruleSL,
                XmNheight, 30,
                XmNwidth, 150,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, ruleLabel1,
                XmNmarginHeight, 0,
                NULL );
    ruleLabel2 = XtVaCreateWidget( "Command: ",
                xmLabelWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, ruleSL,
                XmNheight, 30,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, ruleSpecTF,
                NULL );
    ruleCmdTF = XtVaCreateWidget( "rulePixmapTF",
                xmTextFieldWidgetClass, rulesForm,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, ruleSL,
                XmNheight, 30,
                XmNleftAttachment, XmATTACH_WIDGET,
                XmNleftWidget, ruleLabel2,
                XmNrightAttachment, XmATTACH_FORM,
                XmNrightOffset, 10,
                XmNmarginHeight, 0,
                NULL );

    XmFontListFree( fontList );
  }

  XmTextFieldSetString( defaultCmdTF, defaultCommand );
  loadDevicesData( devSL );
  loadRules( ruleSL, currentRuleTypeIndex );

  isPoppedUp = True;
  XtPopup( appResShell, XtGrabNone );
  XRaiseWindow( XtDisplay(appResShell), XtWindow(appResShell) );
}

void popupBPChooser( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  if( !bpChooser )
  {
    Widget parent, rc, da;
    bpChooser = XmCreateFileSelectionDialog( w, "bpChooser", NULL, 0 );
    parent = XtParent( XmFileSelectionBoxGetChild( bpChooser, 
	XmDIALOG_SEPARATOR ) );
    rc = XtVaCreateManagedWidget( "rc",
                xmRowColumnWidgetClass, parent,
                NULL );
    da = XtVaCreateManagedWidget( "da",
                xmDrawnButtonWidgetClass, rc,
                XmNlabelType, XmPIXMAP,
		XmNheight, 100,
		XmNrecomputeSize, False,
                NULL );
    XtVaSetValues( XmFileSelectionBoxGetChild(bpChooser,XmDIALOG_HELP_BUTTON),
        XtVaTypedArg, XmNlabelString, XmRString, "Apply", 6, NULL );
    XtAddCallback( bpChooser, XmNhelpCallback, bpChooserApplyCB, (XtPointer)da );
    XtAddCallback( bpChooser, XmNcancelCallback, bpChooserCancelCB, NULL );
    XtAddCallback( bpChooser, XmNokCallback, bpChooserOKCB, (XtPointer)x1 );
  }
  XtManageChild( bpChooser );
}

void bpChooserCancelCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  XtUnmanageChild( bpChooser );
}

void bpChooserApplyCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)x2;
  Widget da = (Widget)x1;
  char *filename;
  Pixel fg, bg;
  Pixmap p;

  XtVaGetValues( da, XmNforeground, &fg, XmNbackground, &bg, NULL );

  XmStringGetLtoR( cbs->value, XmSTRING_DEFAULT_CHARSET, &filename );
  if( (p = loadPixmap( da, filename, fg, bg )) == XmUNSPECIFIED_PIXMAP )
  {
    char errorString[256];
    sprintf( errorString, "Cannot Load Pixmap %s", filename );
    popupErrorDialog( errorString );
    return;
  }
  else
    XtVaSetValues( da, XmNlabelPixmap, p, NULL );

  XtFree( filename );
}

void bpChooserOKCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)x2;
  Widget tf = (Widget)x1;
  char *filename;

  XmStringGetLtoR( cbs->value, XmSTRING_DEFAULT_CHARSET, &filename );
  XmTextFieldSetString( tf, filename );

  XtFree( filename );
  XtUnmanageChild( bpChooser );
}

void ruleTypeCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  currentRuleType = w;

  if( w == rulesPB[_FILENAME] )
  {
    XtVaSetValues( ruleLabel1, XtVaTypedArg, XmNlabelString,
	XmRString, "Filename: ", 11, NULL );
    if( XtIsManaged(ruleCmdTF) )  XtUnmanageChild( ruleCmdTF );
    if( XtIsManaged(ruleLabel2) )  XtUnmanageChild( ruleLabel2 );
    currentRuleTypeIndex = 0;
  }
  else if( w == rulesPB[_FILE_COMM] )
  {
    XtVaSetValues( ruleLabel1, XtVaTypedArg, XmNlabelString,
        XmRString, "Filename: ", 11, NULL );
    if( !XtIsManaged(ruleCmdTF) )  XtManageChild( ruleCmdTF );
    if( !XtIsManaged(ruleLabel2) )  XtManageChild( ruleLabel2 );
    currentRuleTypeIndex = 1;
  }
  else if( w == rulesPB[_EXTENSION] )
  {
    XtVaSetValues( ruleLabel1, XtVaTypedArg, XmNlabelString,
        XmRString, "Extension: ", 12, NULL );
    if( XtIsManaged(ruleCmdTF) )  XtUnmanageChild( ruleCmdTF );
    if( XtIsManaged(ruleLabel2) )  XtUnmanageChild( ruleLabel2 );
    currentRuleTypeIndex = 2;
  }
  else if( w == rulesPB[_EXT_COMMAND] )
  {
    XtVaSetValues( ruleLabel1, XtVaTypedArg, XmNlabelString,
        XmRString, "Extension: ", 12, NULL );
    if( !XtIsManaged(ruleCmdTF) )  XtManageChild( ruleCmdTF );
    if( !XtIsManaged(ruleLabel2) )  XtManageChild( ruleLabel2 );
    currentRuleTypeIndex = 3;
  }
  else if( w == rulesPB[_FILESTRING] )
  {
    XtVaSetValues( ruleLabel1, XtVaTypedArg, XmNlabelString,
        XmRString, "Magic String: ", 15, NULL );
    if( XtIsManaged(ruleCmdTF) )  XtUnmanageChild( ruleCmdTF );
    if( XtIsManaged(ruleLabel2) )  XtUnmanageChild( ruleLabel2 );
    currentRuleTypeIndex = 4;
  }
  else if( w == rulesPB[_FILESTR_COMM] )
  {
    XtVaSetValues( ruleLabel1, XtVaTypedArg, XmNlabelString,
        XmRString, "Magic String: ", 15, NULL );
    if( !XtIsManaged(ruleCmdTF) )  XtManageChild( ruleCmdTF );
    if( !XtIsManaged(ruleLabel2) )  XtManageChild( ruleLabel2 );
    currentRuleTypeIndex = 5;
  }

  loadRules( ruleSL, currentRuleTypeIndex );
}

void loadRules( w, i )
  Widget w;
  int i;
{
  int k, j, num = numRules(rules), max1 = 0, max2 = 0;
  char buffer[256];

  for( k = 0; k < num; k++ )
  {
    struct Rule *r = getIthRule( rules, k );
    if( !strcmp( ruleTypes[i], r->type ) )
    {
      if( strlen(r->pixmapName) > max1 )  max1 = strlen(r->pixmapName);
      if( strlen(r->name) > max2 )  max2 = strlen(r->name);
    } 
  }

  max1 += 3;
  max2 += 3;

  XmListDeleteAllItems( w );
  for( k = 0; k < num; k++ )
  {
    struct Rule *r = getIthRule( rules, k );
    if( !strcmp( ruleTypes[i], r->type ) )
    {  
      XmString xms;
      strcpy( buffer, r->pixmapName );
      for( j = strlen(r->pixmapName); j < max1; j++ )  strcat( buffer, " " );
      strcat( buffer, r->name );
      if( i == _FILE_COMM || i == _EXT_COMMAND || i == _FILESTR_COMM )
      {
        for( j = strlen(r->name); j < max2; j++ )  strcat( buffer, " " );
        strcat( buffer , "\"" );
        strcat( buffer, r->command );
        strcat( buffer , "\"" );
      }
      xms = XmStringCreateLtoR( buffer, "set1" );
      XmListAddItemUnselected( w, xms, 0 );
      XmStringFree( xms );
    }
  }
}

void appResApplyCB( w, x1, x2 )
  Widget w; 
  XtPointer x1, x2;
{
  char *defaultLaunchCommand;

  defaultLaunchCommand = XmTextFieldGetString( defaultCmdTF );
  strcpy( defaultCommand, defaultLaunchCommand );
  XtFree( defaultLaunchCommand );
}

void insertDevCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  insertDev( devLabelTF, devPixmapTF, devCmdTF );
  loadDevicesData( devSL );
}

void insertRuleCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  char *name, *command, *pixmapName;
  Pixmap p;
  Pixel fg, bg;
  struct Rule *r;

  XtVaGetValues( XtParent(w), XmNforeground, &fg,
        XmNbackground, &bg, NULL );

  name = XmTextFieldGetString( ruleSpecTF );
  pixmapName = XmTextFieldGetString( rulePixmapTF );

  if( XtIsManaged(ruleCmdTF) )  
  {
    command = XmTextFieldGetString( ruleCmdTF );
    if( !strcmp( name, "" ) || !strcmp( pixmapName, "" ) || !strcmp( command, "" ) )
    {
      popupErrorDialog( "All Rule Fields Must Have Values..." );
      XtFree( name );
      XtFree( command );
      XtFree( pixmapName );
      return;
    }
  }
  else
  {
    if( !strcmp( name, "" ) || !strcmp( pixmapName, "" ) )
    {
      popupErrorDialog( "All Rule Fields Must Have Values..." );
      XtFree( name );
      XtFree( pixmapName );
      return;
    }
  }

  if( (p = loadPixmap( XtParent(w), pixmapName, fg, bg )) == XmUNSPECIFIED_PIXMAP )
  {
    char errorString[256];
    sprintf( errorString, "Cannot Load Pixmap %s", pixmapName );
    popupErrorDialog( errorString );
    return;
  }

  r = appendRule( &rules );
  strcpy( r->pixmapName, pixmapName );
  strcpy( r->name, name );
  strcpy( r->type, ruleTypes[currentRuleTypeIndex] );
  r->pixmap = p;

  XmTextFieldSetString( ruleSpecTF, "" );
  XmTextFieldSetString( rulePixmapTF, "" );

  if( XtIsManaged(ruleCmdTF) )
  {
    strcpy( r->command, command );
    XmTextFieldSetString( ruleCmdTF, "" );
  }
  else
    strcpy( r->command, "" );

  XtFree( name );
  XtFree( command );
  XtFree( pixmapName );

  loadRules( ruleSL, currentRuleTypeIndex );
}

void deleteDevCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  int *pos_list, pos_count;

  if( XmListGetSelectedPos( devSL, &pos_list, &pos_count ) )
  {
    deleteDevice( pos_list[0]-1 );
    XtFree( (char *)pos_list );
    loadDevicesData( devSL );
  }
  else
    popupErrorDialog( "No Device Is Chosen..." );
}

void deleteRuleCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  int *pos_list, pos_count, j, count = 0, num = numRules(rules);

  if( XmListGetSelectedPos( ruleSL, &pos_list, &pos_count ) )
  {
    for( j = 0; j < num; j++ )
    {
      struct Rule *r = getIthRule( rules, j );
      if( !strcmp( ruleTypes[currentRuleTypeIndex], r->type ) )
      {
        count++;
        if( count == pos_list[0] )  
        {
	  removeIthRule( &rules, j );
	  break;
	}
      }
    }
    XtFree( (char *)pos_list );
    loadRules( ruleSL, currentRuleTypeIndex );
  }
  else
    popupErrorDialog( "No Rule Is Chosen..." );
}

void appResSaveCB( w, x1, x2 )
  Widget w; 
  XtPointer x1, x2;
{
  FILE *fd;
  char buffer[256], *defaultLaunchCommand;
  int k, num = numRules(rules);

  sprintf( buffer, "%s/%s", getenv("HOME"), APP_RES_FILENAME );
  if( (fd = fopen(buffer,"w")) == (FILE *)NULL )
  {
    char errorString[256];
    sprintf( errorString, "Cannot Open %s", buffer );
    popupErrorDialog( errorString );
    return;
  }

  defaultLaunchCommand = XmTextFieldGetString( defaultCmdTF );
  strcpy( defaultCommand, defaultLaunchCommand );
  XtFree( defaultLaunchCommand );
  fprintf( fd, "mfm*defaultLaunchCommand: %s\n", defaultCommand );

  if( num > 0 )  fprintf( fd, "mfm*rules: %s\n", "\\" );
  for( k = 0; k < num; k++ )
  {
    struct Rule *r = getIthRule( rules, k );
    if( !strcmp( r->type, ruleTypes[_FILESTRING] ) ||
	!strcmp( r->type, ruleTypes[_FILESTR_COMM] ) )
    {
      fprintf( fd, "\t%s\t%s\t\"%s\"", r->type, r->pixmapName, r->name );
    }
    else
    {
      fprintf( fd, "\t%s\t%s\t%s", r->type, r->pixmapName, r->name );
    }
    if( !strcmp( r->type, ruleTypes[_EXT_COMMAND] ) ||
        !strcmp( r->type, ruleTypes[_FILE_COMM] ) ||
        !strcmp( r->type, ruleTypes[_FILESTR_COMM] ) )
    {
      fprintf( fd, "\t\"%s\"", r->command );
    }

    if( k != num-1 )  
	fprintf( fd, " %s%s%s\n", "\\", "n", "\\" );
    else
	fprintf( fd, "\n" );
  }

  writeDeviceResources( fd );

  fclose( fd );
}

void appResDoneCB( w, x1, x2 )
  Widget w; 
  XtPointer x1, x2;
{
  isPoppedUp = False;
  XtPopdown( appResShell );
}

void readApplicationResources( toplevel )
  Widget toplevel;
{
  XrmDatabase resDB;
  char *ptr = getenv("EDITOR"), buffer[256];
  
  sprintf( buffer, "%s/%s", getenv("HOME"), APP_RES_FILENAME );
  XrmInitialize();
  resDB = XrmGetDatabase( XtDisplay(toplevel) );
  XrmCombineFileDatabase( buffer, &resDB, True );

  strcpy( editorCommand, "" );
  if( ptr )  strcpy( editorCommand, ptr );
  if( !strcmp( editorCommand, "vi" ) )
	strcpy( editorCommand, "xterm -e vi" );

  XtVaGetApplicationResources( toplevel, &appResData, resources,
	XtNumber(resources), NULL );
}

void parseApplicationResources( w, fg, bg )
  Widget w;
  Pixel fg, bg;
{
  strcpy( defaultCommand, appResData.defaultLaunchCommand );
  loadSpecialPixmaps( w, fg, bg );
  parseRules( appResData.rules, w, fg, bg );
}

void parseRules( rulesString, w, fg, bg )
  char *rulesString;
  Widget w;
  Pixel fg, bg;
{
  char *ptr = rulesString;
  char buffer[NAME_LENGTH];

  while( sscanf( ptr, "%[^\n]\n", buffer ) == 1 )
  {
      struct Rule *r;
      char temp1[NAME_LENGTH], type[15];
      Pixmap p;
      int i;

      if( sscanf( buffer, "%s %s", type, temp1 ) != 2 )  continue;

      for( i = 0; i < numRuleTypes; i++ )
          if( !strcmp( type, ruleTypes[i] ) )  break;
      if( i >= numRuleTypes )  continue;

      if( (p = loadPixmap( w, temp1, fg, bg )) == XmUNSPECIFIED_PIXMAP )
	 continue;

      r = appendRule( &rules );
      strcpy( r->pixmapName, temp1 );
      strcpy( r->type, type );
      r->pixmap = p;

      strcpy( r->name, "" );
      if( !strcmp( type, ruleTypes[_FILENAME] ) || 
	  !strcmp( type, ruleTypes[_FILE_COMM] ) ||
	  !strcmp( type, ruleTypes[_EXTENSION] ) ||
	  !strcmp( type, ruleTypes[_EXT_COMMAND] ) )
      {
	sscanf( buffer, "%*s %*s %s", temp1 );
	strcpy( r->name, temp1 );
      }
      else if( !strcmp( type, ruleTypes[_FILESTRING] ) || 
	       !strcmp( type, ruleTypes[_FILESTR_COMM] ) )
      {
	char *ptr;
	strcpy( r->name, strchr( buffer, '\"' ) + 1 );
 	ptr = strchr( r->name, '\"' );
	*ptr = '\0';
      }

      strcpy( r->command, "" );
      if( !strcmp( type, ruleTypes[_EXT_COMMAND] ) || 
	  !strcmp( type, ruleTypes[_FILE_COMM] ) ||
	  !strcmp( type, ruleTypes[_FILESTR_COMM] ) )
      {
	char *ptr;
	if( !strcmp( type, ruleTypes[_FILESTR_COMM] ) )
	{
	  ptr = strchr( buffer, '\"' );
	  *ptr = ' ';
	  ptr = strchr( buffer, '\"' );
          *ptr = ' ';
	}

	strcpy( r->command, strchr( buffer, '\"' ) + 1 );
	ptr = strrchr( r->command, '\"' );
	*ptr = '\0';
      }

      if( strlen(buffer)+1 >= strlen(ptr) )
	break;
      else
	ptr += strlen( buffer ) + 1;

      memset( (void *)buffer, '\0', sizeof(buffer) );
  }
}

Pixmap isaDevice( name )
  char *name;
{
  if( !strcmp( name, "PRINTER" ) )  return specialTypes.PRINTER;
  if( !strcmp( name, "FLOPPY" ) )  return specialTypes.FLOPPY;
  if( !strcmp( name, "TAPES" ) )  return specialTypes.TAPES;
  return XmUNSPECIFIED_PIXMAP;
}

Pixmap getPixmap( name, s, type, perm )
  char *name;
  struct stat s;
  int type;
  char *perm;
{
  int k, num = numRules(rules);
  char *ptr;

  if( !strcmp( name, "MULTIPLE" ) )  return specialTypes.MULTIPLE;
  if( !strcmp( name, "/" ) )  return specialTypes.DIRECTORY;
  if( !strcmp( name, "BLANK" ) )  return specialTypes.BLANK;
  if( !strcmp( name, "PRINTER" ) )  return specialTypes.PRINTER;
  if( !strcmp( name, "FLOPPY" ) )  return specialTypes.FLOPPY;
  if( !strcmp( name, "TAPES" ) )  return specialTypes.TAPES;
  if( !strcmp( name, "HOME" ) || !strcmp( name, getenv("HOME") ) )  
	return specialTypes.HOME;

  if( (ptr = strrchr( name, '/' )) == (char *)NULL )  
	return specialTypes.REG_DEFAULT;
  else
        ptr++;

  for( k = 0; k < num; k++ )
  {
    struct Rule *r = getIthRule( rules, k );
    if( !strcmp( r->type, ruleTypes[_FILENAME] ) &&
	!strcmp( ptr, r->name ) )
    {
      return r->pixmap;
    }
    else if( !strcmp( r->type, ruleTypes[_FILE_COMM] ) &&
        !strcmp( ptr, r->name ) )
    {
      return r->pixmap;
    }
    else if( ( !strcmp( r->type, ruleTypes[_EXTENSION] ) ||
	       !strcmp( r->type, ruleTypes[_EXT_COMMAND] ) ) &&
	     extensionMatches( ptr, r->name ) )
    {
      return r->pixmap;
    }
    else if( !strcmp( r->type, ruleTypes[_FILESTRING] ) ||
             !strcmp( r->type, ruleTypes[_FILESTR_COMM] ) )
    {
      char retBuffer[256];
      if( fileCmd( name, retBuffer ) != -1 && 
	  !strcmp( retBuffer, r->name ) )
      {
	return r->pixmap;
      }
    }
  }

  if( type == LOCAL_FILESYSTEM )
  {
    if( S_ISDIR( s.st_mode ) )
      return specialTypes.DIRECTORY;
#ifdef S_ISLNK
    else if( S_ISLNK( s.st_mode ) )
      return specialTypes.LINK;
#endif
    else if( S_ISCHR( s.st_mode ) )
      return specialTypes.CHAR_SPECIAL;
    else if( S_ISBLK( s.st_mode ) )
      return specialTypes.BLK_SPECIAL;
    else if( S_ISFIFO( s.st_mode ) )
      return specialTypes.FIFO;
#ifdef S_ISSOCK
    else if( S_ISSOCK( s.st_mode ) )
      return specialTypes.SOCKET;
#endif
    else if( IS_EXECUTABLE( s.st_mode ) )
      return specialTypes.EXECUTABLE;
    else
      return specialTypes.REG_DEFAULT;
  }
  else
  {
    if( perm[0] == 'l' )
      return specialTypes.LINK;
    else if( perm[0] == 'd' )
      return specialTypes.DIRECTORY;
    else if( perm[3] == 'x' || perm[6] == 'x' || perm[9] == 'x' )
      return specialTypes.EXECUTABLE;
    else 
      return specialTypes.REG_DEFAULT;
  }
}

Boolean extensionMatches( name, ext )
  char *name, *ext;
{
  int i, se, sn;

  se = strlen(ext);
  sn = strlen(name);
  if( se > sn )  return False;

  for( i = 0; i < se; i++ )
  {
    if( ext[i] != name[sn-se+i] )
	return False;
  }

  return True;
}

int getDefaultCommand( name, cmd )
  char *name, *cmd;
{
  if( strcmp( defaultCommand, "" ) )
  {
    strcpy( cmd, defaultCommand );
    insertFileName( name, cmd );
    return 0;
  }
  return 1;
}

void getCommand( name, cmd )
  char *name, *cmd;
{
  int i, num = numRules(rules);
  char *ptr;

  memset( (void *)cmd, '\0', sizeof(cmd) );

  if( (ptr = strrchr( name, '/' )) == (char *)NULL )  
  {
    strcpy( cmd, name );
    return;
  }
  ptr++;

  for( i = 0; i < num; i++ )
  {
    struct Rule *r = getIthRule( rules, i );
    if( !strcmp( r->type, ruleTypes[_EXT_COMMAND] ) &&
	extensionMatches( ptr, r->name ) )
    {
      strcpy( cmd, r->command );
      insertFileName( name, cmd );
      return;
    }
    else if( !strcmp( r->type, ruleTypes[_FILESTR_COMM] ) )
    {
      char retBuffer[256];
      if( fileCmd( name, retBuffer ) != -1 &&
          !strcmp( retBuffer, r->name ) )
      {
        strcpy( cmd, r->command );
        insertFileName( name, cmd );
        return;
      }
    }
    else if( !strcmp( r->type, ruleTypes[_FILE_COMM] ) &&
	!strcmp( ptr, r->name ) )
    {
      strcpy( cmd, r->command );
      insertFileName( name, cmd );
      return;
    }
  }

  strcpy( cmd, name );
  return;
}

Pixmap loadPixmap( w, name, fg, bg )
  Widget w;
  char* name;
  Pixel fg, bg;
{
  char buffer[256];
  Pixmap p = (Pixmap)NULL;

#ifdef HAVE_XPMLIB

  XpmAttributes attrib;
  XpmColorSymbol transColor[1];
  Pixmap mask = (Pixmap)NULL;

  transColor[0].name = NULL;
  transColor[0].value = "none";
  transColor[0].pixel = bg;

  attrib.valuemask = XpmColorSymbols | XpmCloseness;
  attrib.closeness = 65535;
  attrib.colorsymbols = transColor;
  attrib.numsymbols = 1;
  if( XpmReadFileToPixmap( XtDisplay(w), DefaultRootWindow(XtDisplay(w)),
        name, &p, &mask, &attrib ) == XpmSuccess )
  {
    XpmFreeAttributes( &attrib );
    if( mask )  XFreePixmap( XtDisplay(w), mask );
    return p;
  }
  XpmFreeAttributes( &attrib );
  if( mask )  XFreePixmap( XtDisplay(w), mask );

  sprintf( buffer, "%s/%s%s", XBMPATH, "mfm/", name );
  attrib.valuemask = XpmColorSymbols | XpmCloseness;
  attrib.closeness = 65535;
  attrib.colorsymbols = transColor;
  attrib.numsymbols = 1;
  if( XpmReadFileToPixmap( XtDisplay(w), DefaultRootWindow(XtDisplay(w)),
        buffer, &p, &mask, &attrib ) == XpmSuccess )
  {
    XpmFreeAttributes( &attrib );
    if( mask )  XFreePixmap( XtDisplay(w), mask );
    return p;
  }
  XpmFreeAttributes( &attrib );
  if( mask )  XFreePixmap( XtDisplay(w), mask );

  sprintf( buffer, "%s%s", "mfm/", name );
  attrib.valuemask = XpmColorSymbols | XpmCloseness;
  attrib.closeness = 65535;
  attrib.colorsymbols = transColor;
  attrib.numsymbols = 1;
  if( XpmReadFileToPixmap( XtDisplay(w), DefaultRootWindow(XtDisplay(w)),
        buffer, &p, &mask, &attrib ) == XpmSuccess )
  {
    XpmFreeAttributes( &attrib );
    if( mask )  XFreePixmap( XtDisplay(w), mask );
    return p;
  }
  XpmFreeAttributes( &attrib );
  if( mask )  XFreePixmap( XtDisplay(w), mask );

#endif

  if( (p = XmGetPixmap( XtScreen(w), name, fg, bg )) != XmUNSPECIFIED_PIXMAP )
    return p;

  sprintf( buffer, "%s%s", "mfm/", name );
  if( (p = XmGetPixmap( XtScreen(w), buffer, fg, bg )) != XmUNSPECIFIED_PIXMAP )
    return p;
  
  sprintf( buffer, "%s/%s%s", XBMPATH, "mfm/", name );
  if( (p = XmGetPixmap( XtScreen(w), buffer, fg, bg )) != XmUNSPECIFIED_PIXMAP )
    return p;

  fprintf( stderr, "mfm: cannot find pixmap %s\n", name );
  return p;
}

void loadIncludedPixmap( display, window, bits, p, bg, name )
  Display *display;
  Window window;
  char **bits;
  Pixmap *p;
  Pixel bg;
  char *name;
{
#ifdef HAVE_XPMLIB
  Pixmap mask = (Pixmap)NULL;
  XpmColorSymbol transColor[1];
  XpmAttributes attrib;

  transColor[0].name = NULL;
  transColor[0].value = "none";
  transColor[0].pixel = bg;

  attrib.valuemask = XpmReturnPixels | XpmColorSymbols;
  attrib.colorsymbols = transColor;
  attrib.numsymbols = 1;

  if( XpmCreatePixmapFromData( display, window, bits, p, &mask,
					&attrib ) != XpmSuccess )
    fprintf( stderr, "XpmCreatePixmapFromData() failed for %s...\n", name );
  XpmFreeAttributes( &attrib );
  if( mask )  XFreePixmap( display, mask );
#endif
}

void loadSpecialPixmaps( w, fg, bg )
  Widget w;
  Pixel fg, bg;
{
  Display *display = XtDisplay( w );
  Window window = DefaultRootWindow( display );
  int depth = DefaultDepthOfScreen( XtScreen(w) );

#ifdef HAVE_XPMLIB

  loadIncludedPixmap( display, window, plain_folder_xpm, 
	&(specialTypes.DIRECTORY), bg, "DIRECTORY" );
  loadIncludedPixmap( display, window, network_xpm,
	&(specialTypes.SOCKET), bg, "SOCKET" );
  loadIncludedPixmap( display, window, plain_doc_xpm,
	&(specialTypes.FIFO), bg, "FIFO" );
  loadIncludedPixmap( display, window, plain_doc_xpm,
	&(specialTypes.REG_DEFAULT), bg, "REG_DEFAULT" );
  loadIncludedPixmap( display, window, exec_xpm,
	&(specialTypes.EXECUTABLE), bg, "EXECUTABLE" );
  loadIncludedPixmap( display, window, multiple_xpm,
	&(specialTypes.MULTIPLE), bg, "MULTIPLE" );
  loadIncludedPixmap( display, window, house_xpm,
	&(specialTypes.HOME), bg, "HOME" );
  loadIncludedPixmap( display, window, link_dir_xpm,
	&(specialTypes.LINK), bg, "LINK" );
  loadIncludedPixmap( display, window, printer_xpm,
	&(specialTypes.PRINTER), bg, "PRINTER" );
  loadIncludedPixmap( display, window, floppy_xpm,
	&(specialTypes.FLOPPY), bg, "FLOPPY" );
  loadIncludedPixmap( display, window, tapes_xpm,
	&(specialTypes.TAPES), bg, "TAPES" );

#else

  specialTypes.DIRECTORY = XCreatePixmapFromBitmapData( display, window,
	plain_folder_bits, plain_folder_width, plain_folder_height,
	fg, bg, depth );
  specialTypes.SOCKET = XCreatePixmapFromBitmapData( display, window,
	network_bits, network_width, network_height,
	fg, bg, depth );
  specialTypes.FIFO = XCreatePixmapFromBitmapData( display, window,
	plain_doc_bits, plain_doc_width, plain_doc_height,
	fg, bg, depth );
  specialTypes.REG_DEFAULT = XCreatePixmapFromBitmapData( display, window,
	plain_doc_bits, plain_doc_width, plain_doc_height,
	fg, bg, depth );
  specialTypes.EXECUTABLE = XCreatePixmapFromBitmapData( display, window,
	exec_bits, exec_width, exec_height,
	fg, bg, depth );
  specialTypes.MULTIPLE = XCreatePixmapFromBitmapData( display, window,
	multiple_bits, multiple_width, multiple_height,
	fg, bg, depth );
  specialTypes.HOME = XCreatePixmapFromBitmapData( display, window,
	house_bits, house_width, house_height,
	fg, bg, depth );
  specialTypes.LINK = XCreatePixmapFromBitmapData( display, window,
	link_dir_bits, link_dir_width, link_dir_height,
	fg, bg, depth );
  specialTypes.PRINTER = XCreatePixmapFromBitmapData( display, window,
	printer_bits, printer_width, printer_height,
	fg, bg, depth );
  specialTypes.FLOPPY = XCreatePixmapFromBitmapData( display, window,
        floppy_bits, floppy_width, floppy_height,
	fg, bg, depth );
  specialTypes.TAPES = XCreatePixmapFromBitmapData( display, window,
        tapes_bits, tapes_width, tapes_height,
	fg, bg, depth );

#endif

  specialTypes.BLANK = XCreatePixmapFromBitmapData( display, window,
        blank_bits, blank_width, blank_height,
        fg, bg, depth );
  specialTypes.CHAR_SPECIAL = XCreatePixmapFromBitmapData( display, window,
        char_special_bits, char_special_width, char_special_height,
        fg, bg, depth );
  specialTypes.BLK_SPECIAL = XCreatePixmapFromBitmapData( display, window,
        block_special_bits, block_special_width, block_special_height,
        fg, bg, depth );
}

int fileCmd( filename, returnBuffer )
  char *filename, *returnBuffer;
{
  FILE *in;
  char cmd[256], *ptr;
  int n;

  memset( (void *)returnBuffer, '\0', sizeof(returnBuffer) );
  sprintf( cmd, "file %s", filename );

  if( (in = popen( cmd, "r" )) == (FILE *)NULL )
  {
    sprintf( returnBuffer, "popen() failed for: %s", cmd );
    return -1;
  }

  fgets( cmd, sizeof(cmd), in );
  if( (ptr = strchr( cmd, '\t' )) != NULL )  strcpy( returnBuffer, ptr+1 );
  if( (ptr = strchr( returnBuffer, '\n' )) != NULL )  *ptr = '\0';

  pclose( in );
  return 0;
}

void insertFileName( file, cmd )
  char *file, *cmd;
{
  char *ptr, buffer[1056];
  int j = 0;

  if( !strcmp( file, "" ) )
  {
    return;
  }
  else if( !strcmp( cmd, "" ) )
  {
    strcpy( cmd, file );
    return;
  }

  strcpy( buffer, "" );

  if( (ptr = strstr( cmd, "$ALL" )) != NULL )
  {
    strncpy( buffer, cmd, ptr-cmd );
    buffer[ptr-cmd] = '\0';
    strcat( buffer, file );
    ptr += 4;
    strcat( buffer, ptr );
    strcpy( cmd, buffer );
    j = 1;
  }
  if( (ptr = strstr( cmd, "$PATH+FILENAME" )) != NULL )
  {
    strncpy( buffer, cmd, ptr-cmd );
    buffer[ptr-cmd] = '\0';
    strcat( buffer, file );
    ptr += 14;
    strcat( buffer, ptr );
    strcpy( cmd, buffer );
    j = 1;
  }
  if( (ptr = strstr( cmd, "$FILENAME" )) != NULL )
  {
    strncpy( buffer, cmd, ptr-cmd );
    buffer[ptr-cmd] = '\0';
    strcat( buffer, strrchr(file,'/')+1 );
    ptr += 9;
    strcat( buffer, ptr );
    strcpy( cmd, buffer );
    j = 1;
  }
  if( (ptr = strstr( cmd, "$PATH" )) != NULL )
  {
    char *sptr;
    strncpy( buffer, cmd, ptr-cmd );
    buffer[ptr-cmd] = '\0';
    sptr = strrchr( file, '/' );
    strncat( buffer, file, sptr-file );
    ptr += 5;
    strcat( buffer, ptr );
    strcpy( cmd, buffer );
    j = 1;
  }
  if( (ptr = strstr( cmd, "$EDITOR" )) != NULL )
  {
    if( !strcmp( editorCommand, "" ) )
    {
      strcpy( cmd, "" );
      popupErrorDialog( "EDITOR environment variable is not set..." );
      return;
    }
    else
    {
      strncpy( buffer, cmd, ptr-cmd );
      buffer[ptr-cmd] = '\0';
      strcat( buffer, editorCommand );
      ptr += 7;
      strcat( buffer, ptr );
      strcpy( cmd, buffer );
    }
  }

  if( j == 0 )
  {
    strcat( cmd, " " );
    strcat( cmd, file );
  }
}

void setToolBarPixmaps( duplicateWidget, deleteWidget, attribWidget,
	createDirWidget, ftpWidget, newFMWidget, infoWidget )
  Widget duplicateWidget, deleteWidget, attribWidget, 
	 createDirWidget, ftpWidget, newFMWidget, infoWidget;
{
  Display *display = XtDisplay( duplicateWidget );
  Window window = DefaultRootWindow( display );
  int depth = DefaultDepthOfScreen( XtScreen(duplicateWidget) );
  Pixel fg, bg;
  Pixmap p;

  XtVaGetValues( duplicateWidget, XmNforeground, &fg, XmNbackground, &bg, NULL );

#ifdef HAVE_XPMLIB

  loadIncludedPixmap( display, window, duplicate_xpm, &p, bg, "DUPLICATE" );
  XtVaSetValues( duplicateWidget, XmNlabelPixmap, p, NULL );
  loadIncludedPixmap( display, window, delete_xpm, &p, bg, "DELETE" );
  XtVaSetValues( deleteWidget, XmNlabelPixmap, p, NULL );
  loadIncludedPixmap( display, window, attributes_xpm, &p, bg, "ATTRIBUTES" );
  XtVaSetValues( attribWidget, XmNlabelPixmap, p, NULL );
  loadIncludedPixmap( display, window, createDir_xpm, &p, bg, "CREATE DIR" );
  XtVaSetValues( createDirWidget, XmNlabelPixmap, p, NULL );
#if MAJOR_VERSION >= 2
  loadIncludedPixmap( display, window, ftp_xpm, &p, bg, "FTP" );
  XtVaSetValues( ftpWidget, XmNlabelPixmap, p, NULL );
#endif
  loadIncludedPixmap( display, window, filing_xpm, &p, bg, "NEW FILEMANAGER" );
  XtVaSetValues( newFMWidget, XmNlabelPixmap, p, NULL );
  loadIncludedPixmap( display, window, question_xpm, &p, bg, "QUESTION" );
  XtVaSetValues( infoWidget, XmNlabelPixmap, p, NULL );

#else

  p = XCreatePixmapFromBitmapData( display, window,
        duplicate_bits, duplicate_width, duplicate_height, fg, bg, depth );
  XtVaSetValues( duplicateWidget, XmNlabelPixmap, p, NULL );
  p = XCreatePixmapFromBitmapData( display, window,
        delete_bits, delete_width, delete_height, fg, bg, depth );
  XtVaSetValues( deleteWidget, XmNlabelPixmap, p, NULL );
  p = XCreatePixmapFromBitmapData( display, window,
        attributes_bits, attributes_width, attributes_height, fg, bg, depth );
  XtVaSetValues( attribWidget, XmNlabelPixmap, p, NULL );
  p = XCreatePixmapFromBitmapData( display, window,
        createDir_bits, createDir_width, createDir_height, fg, bg, depth );
  XtVaSetValues( createDirWidget, XmNlabelPixmap, p, NULL );
#if MAJOR_VERSION >= 2
  p = XCreatePixmapFromBitmapData( display, window,
	ftp_bits, ftp_width, ftp_height, fg, bg, depth );
  XtVaSetValues( ftpWidget, XmNlabelPixmap, p, NULL );
#endif
  p = XCreatePixmapFromBitmapData( display, window,
        filing32_bits, filing32_width, filing32_height, fg, bg, depth );
  XtVaSetValues( newFMWidget, XmNlabelPixmap, p, NULL );
  p = XCreatePixmapFromBitmapData( display, window,
	question_bits, question_width, question_height, fg, bg, depth );
  XtVaSetValues( infoWidget, XmNlabelPixmap, p, NULL );

#endif
}
